import json
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any, cast

import yaml
from cognite.client import CogniteClient
from cognite.client.data_classes import (
    AssetWrite,
    LabelDefinitionWrite,
    RelationshipWrite,
)
from cognite.client.data_classes.capabilities import (
    AssetsAcl,
    Capability,
    RelationshipsAcl,
)
from cognite.client.exceptions import CogniteAPIError, CogniteDuplicatedError

from cognite.neat._graph._tracking.base import Tracker
from cognite.neat._graph._tracking.log import LogTracker
from cognite.neat._issues import IssueList, NeatError, NeatIssue
from cognite.neat._issues.errors import ResourceCreationError, ResourceNotFoundError
from cognite.neat._rules._constants import EntityTypes
from cognite.neat._rules.analysis._asset import AssetAnalysis
from cognite.neat._rules.models import AssetRules
from cognite.neat._rules.models.entities import ClassEntity
from cognite.neat._store import NeatGraphStore
from cognite.neat._utils.auxiliary import create_sha256_hash
from cognite.neat._utils.upload import UploadResult

from ._base import _END_OF_CLASS, CDFLoader


class AssetLoader(CDFLoader[AssetWrite]):
    """Load Assets and their relationships from NeatGraph to Cognite Data Fusions.

    Args:
        graph_store (NeatGraphStore): The graph store to load the data into.
        rules (AssetRules): The rules to load the assets with.
        data_set_id (int): The CDF data set id to load the Assets into.
        use_orphanage (bool): Whether to use an orphanage for assets that are not part
                              of the hierarchy. Defaults to False.
        use_labels (bool): Whether to use labels for assets. Defaults to False.
        external_id_prefix (str | None): The prefix to use for the external ids. Defaults to None.
        create_issues (Sequence[NeatIssue] | None): A list of issues that occurred during reading. Defaults to None.
        tracker (type[Tracker] | None): The tracker to use. Defaults to None.
    """

    def __init__(
        self,
        graph_store: NeatGraphStore,
        rules: AssetRules,
        data_set_id: int,
        use_orphanage: bool = False,
        use_labels: bool = False,
        external_id_prefix: str | None = None,
        create_issues: Sequence[NeatIssue] | None = None,
        tracker: type[Tracker] | None = None,
    ):
        super().__init__(graph_store)

        self.rules = rules
        self.data_set_id = data_set_id

        self.use_labels = use_labels

        self.orphanage = (
            AssetWrite.load(
                {
                    "dataSetId": self.data_set_id,
                    "externalId": (f"{external_id_prefix or ''}orphanage-{data_set_id}" if use_orphanage else None),
                    "name": "Orphanage",
                    "description": "Orphanage for assets whose parents do not exist",
                }
            )
            if use_orphanage
            else None
        )

        self.external_id_prefix = external_id_prefix

        self.processed_assets: set[str] = set()
        self._issues = IssueList(create_issues or [])
        self._tracker: type[Tracker] = tracker or LogTracker

    def _load(self, stop_on_exception: bool = False) -> Iterable[AssetWrite | NeatIssue | type[_END_OF_CLASS]]:
        if self._issues.has_errors and stop_on_exception:
            raise self._issues.as_exception()
        elif self._issues.has_errors:
            yield from self._issues
            return
        if not self.rules:
            # There should already be an error in this case.
            return

        ordered_classes = AssetAnalysis(self.rules).class_topological_sort()

        tracker = self._tracker(
            type(self).__name__,
            [repr(class_.id) for class_ in ordered_classes],
            "classes",
        )

        if self.use_labels:
            yield from self._create_labels()

        if self.orphanage:
            yield self.orphanage
            self.processed_assets.add(cast(str, self.orphanage.external_id))

        yield from self._create_assets(ordered_classes, tracker, stop_on_exception)
        yield from self._create_relationship(ordered_classes, tracker, stop_on_exception)

    def _create_labels(self) -> Iterable[Any]:
        for label in AssetAnalysis(self.rules).define_labels():
            yield LabelDefinitionWrite(name=label, external_id=label, data_set_id=self.data_set_id)
        yield _END_OF_CLASS

    def _create_assets(
        self,
        ordered_classes: list[ClassEntity],
        tracker: Tracker,
        stop_on_exception: bool,
    ) -> Iterable[Any]:
        error: NeatError
        for class_ in ordered_classes:
            tracker.start(repr(class_.id))

            property_renaming_config = AssetAnalysis(self.rules).define_asset_property_renaming_config(class_)

            for identifier, properties in self.graph_store.read(class_.suffix):
                identifier = f"{self.external_id_prefix or ''}{identifier}"

                fields = _process_asset_properties(properties, property_renaming_config)
                # set data set id and external id
                fields["dataSetId"] = self.data_set_id
                fields["externalId"] = identifier

                if self.use_labels:
                    fields["labels"] = [class_.suffix]

                if parent_external_id := fields.get("parentExternalId", None):
                    fields["parentExternalId"] = f"{self.external_id_prefix or ''}{parent_external_id}"

                # check on parent
                if "parentExternalId" in fields and fields["parentExternalId"] not in self.processed_assets:
                    error = ResourceNotFoundError(
                        fields["parentExternalId"],
                        EntityTypes.asset,
                        identifier,
                        EntityTypes.asset,
                        f"Moving the asset {identifier} under orphanage {self.orphanage.external_id}"
                        if self.orphanage
                        else "",
                    )
                    tracker.issue(error)
                    if stop_on_exception:
                        raise error
                    yield error

                    # if orphanage is set asset will use orphanage as parent
                    if self.orphanage:
                        fields["parentExternalId"] = self.orphanage.external_id

                    # otherwise asset will be skipped
                    else:
                        continue

                try:
                    yield AssetWrite.load(fields)
                    self.processed_assets.add(identifier)
                except KeyError as e:
                    error = ResourceCreationError(identifier, EntityTypes.asset, error=str(e))
                    tracker.issue(error)
                    if stop_on_exception:
                        raise error from e
                    yield error

            yield _END_OF_CLASS

    def _create_relationship(
        self,
        ordered_classes: list[ClassEntity],
        tracker: Tracker,
        stop_on_exception: bool,
    ) -> Iterable[Any]:
        for class_ in ordered_classes:
            tracker.start(repr(class_.id))

            property_renaming_config = AssetAnalysis(self.rules).define_relationship_property_renaming_config(class_)

            # class does not have any relationship properties
            if not property_renaming_config:
                continue

            for source_external_id, properties in self.graph_store.read(class_.suffix):
                relationships = _process_relationship_properties(properties, property_renaming_config)

                source_external_id = f"{self.external_id_prefix or ''}{source_external_id}"

                # check if source asset exists
                if source_external_id not in self.processed_assets:
                    error = ResourceCreationError(
                        resource_type=EntityTypes.relationship,
                        identifier=source_external_id,
                        error=(
                            f"Asset {source_external_id} does not exist! "
                            "Aborting creation of relationships which use this asset as the source."
                        ),
                    )
                    tracker.issue(error)
                    if stop_on_exception:
                        raise error
                    yield error
                    continue

                for label, target_external_ids in relationships.items():
                    # we can have 1-many relationships
                    for target_external_id in target_external_ids:
                        target_external_id = f"{self.external_id_prefix or ''}{target_external_id}"
                        # check if source asset exists
                        if target_external_id not in self.processed_assets:
                            error = ResourceCreationError(
                                resource_type=EntityTypes.relationship,
                                identifier=target_external_id,
                                error=(
                                    f"Asset {target_external_id} does not exist! "
                                    f"Cannot create relationship between {source_external_id}"
                                    f" and {target_external_id}. "
                                ),
                            )
                            tracker.issue(error)
                            if stop_on_exception:
                                raise error
                            yield error
                            continue

                        external_id = "relationship_" + create_sha256_hash(f"{source_external_id}_{target_external_id}")
                        try:
                            yield RelationshipWrite(
                                external_id=external_id,
                                source_external_id=source_external_id,
                                target_external_id=target_external_id,
                                source_type="asset",
                                target_type="asset",
                                data_set_id=self.data_set_id,
                                labels=[label] if self.use_labels else None,
                            )
                        except KeyError as e:
                            error = ResourceCreationError(
                                resource_type=EntityTypes.relationship,
                                identifier=external_id,
                                error=str(e),
                            )
                            tracker.issue(error)
                            if stop_on_exception:
                                raise error from e
                            yield error

            yield _END_OF_CLASS

    def _get_required_capabilities(self) -> list[Capability]:
        return [
            AssetsAcl(
                actions=[
                    AssetsAcl.Action.Write,
                    AssetsAcl.Action.Read,
                ],
                scope=AssetsAcl.Scope.DataSet([self.data_set_id]),
            ),
            RelationshipsAcl(
                actions=[
                    RelationshipsAcl.Action.Write,
                    RelationshipsAcl.Action.Read,
                ],
                scope=RelationshipsAcl.Scope.DataSet([self.data_set_id]),
            ),
        ]

    def _upload_to_cdf(
        self,
        client: CogniteClient,
        items: list[AssetWrite] | list[RelationshipWrite] | list[LabelDefinitionWrite],
        dry_run: bool,
        read_issues: IssueList,
    ) -> Iterable[UploadResult]:
        if isinstance(items[0], AssetWrite) and all(isinstance(item, AssetWrite) for item in items):
            yield from self._upload_assets_to_cdf(client, cast(list[AssetWrite], items), dry_run, read_issues)
        elif isinstance(items[0], RelationshipWrite) and all(isinstance(item, RelationshipWrite) for item in items):
            yield from self._upload_relationships_to_cdf(
                client, cast(list[RelationshipWrite], items), dry_run, read_issues
            )
        elif isinstance(items[0], LabelDefinitionWrite) and all(
            isinstance(item, LabelDefinitionWrite) for item in items
        ):
            yield from self._upload_labels_to_cdf(client, cast(list[LabelDefinitionWrite], items), dry_run, read_issues)
        else:
            raise ValueError(f"Item {items[0]} is not supported. This is a bug in neat please report it.")

    def _upload_labels_to_cdf(
        self,
        client: CogniteClient,
        items: list[LabelDefinitionWrite],
        dry_run: bool,
        read_issues: IssueList,
    ) -> Iterable[UploadResult]:
        try:
            created = client.labels.create(items)
        except (CogniteAPIError, CogniteDuplicatedError) as e:
            result = UploadResult[str](name="Label", issues=read_issues)
            result.error_messages.append(str(e))
            result.failed_created.update(item.external_id for item in e.failed + e.unknown)
            result.created.update(item.external_id for item in e.successful)
            yield result
        else:
            for label in created:
                result = UploadResult[str](name="Label", issues=read_issues)
                result.upserted.add(cast(str, label.external_id))
                yield result

    def _upload_assets_to_cdf(
        self,
        client: CogniteClient,
        items: list[AssetWrite],
        dry_run: bool,
        read_issues: IssueList,
    ) -> Iterable[UploadResult]:
        try:
            upserted = client.assets.upsert(items, mode="replace")
        except CogniteAPIError as e:
            result = UploadResult[str](name="Asset", issues=read_issues)
            result.error_messages.append(str(e))
            result.failed_upserted.update(item.external_id for item in e.failed + e.unknown)
            result.upserted.update(item.external_id for item in e.successful)
            yield result
        else:
            for asset in upserted:
                result = UploadResult[str](name="Asset", issues=read_issues)
                result.upserted.add(cast(str, asset.external_id))
                yield result

    def _upload_relationships_to_cdf(
        self,
        client: CogniteClient,
        items: list[RelationshipWrite],
        dry_run: bool,
        read_issues: IssueList,
    ) -> Iterable[UploadResult]:
        try:
            upserted = client.relationships.upsert(items, mode="replace")
        except CogniteAPIError as e:
            result = UploadResult[str](name="Relationship", issues=read_issues)
            result.error_messages.append(str(e))
            result.failed_upserted.update(item.external_id for item in e.failed + e.unknown)
            result.upserted.update(item.external_id for item in e.successful)
            yield result
        else:
            for relationship in upserted:
                result = UploadResult[str](name="relationship", issues=read_issues)
                result.upserted.add(cast(str, relationship.external_id))
                yield result

    def write_to_file(self, filepath: Path) -> None:
        if filepath.suffix not in [".json", ".yaml", ".yml"]:
            raise ValueError(f"File format {filepath.suffix} is not supported")
        dumped: dict[str, list] = {"assets": [], "relationship": []}
        for item in self.load(stop_on_exception=False):
            key = {
                AssetWrite: "assets",
                RelationshipWrite: "relationship",
                NeatIssue: "issues",
                _END_OF_CLASS: "end_of_class",
            }.get(type(item))
            if key is None:
                # This should never happen, and is a bug in neat
                raise ValueError(f"Item {item} is not supported. This is a bug in neat please report it.")
            if key == "end_of_class":
                continue
            dumped[key].append(item.dump())
        with filepath.open("w", encoding=self._encoding, newline=self._new_line) as f:
            if filepath.suffix == ".json":
                json.dump(dumped, f, indent=2)
            else:
                yaml.safe_dump(dumped, f, sort_keys=False)


def _process_asset_properties(properties: dict[str, list[str]], property_renaming_config: dict[str, str]) -> dict:
    metadata: dict[str, str] = {}
    fields: dict[str, str | dict] = {}

    for original_property, values in properties.items():
        if renamed_property := property_renaming_config.get(original_property, None):
            if renamed_property.startswith("metadata."):
                # Asset metadata contains only string values
                metadata[original_property] = ", ".join(values)
            else:
                # Asset fields can contain only one value
                fields[renamed_property] = values[0]

    if metadata:
        fields["metadata"] = metadata

    return fields


def _process_relationship_properties(
    properties: dict[str, list[str]], property_renaming_config: dict[str, str]
) -> dict:
    relationships: dict[str, list[str]] = {}

    for original_property, values in properties.items():
        if renamed_property := property_renaming_config.get(original_property, None):
            relationships[renamed_property] = values

    return relationships



from collections.abc import Set
from datetime import datetime, timezone
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import DataSet, DataSetList
from cognite.client.utils.useful_types import SequenceNotStr
from rdflib import RDF, Literal, Namespace

from cognite.neat._graph.models import Triple

from ._base import DEFAULT_SKIP_METADATA_VALUES, ClassicCDFBaseExtractor, InstanceIdPrefix


class DataSetExtractor(ClassicCDFBaseExtractor[DataSet]):
    """Extract DataSets from Cognite Data Fusions into Neat.

    Args:
        items (Iterable[Asset]): An iterable of assets.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
        to_type (Callable[[Asset], str | None], optional): A function to convert an asset to a type. Defaults to None.
            If None or if the function returns None, the asset will be set to the default type "Asset".
        total (int, optional): The total number of assets to load. If passed, you will get a progress bar if rich
            is installed. Defaults to None.
        limit (int, optional): The maximal number of assets to load. Defaults to None. This is typically used for
            testing setup of the extractor. For example, if you are extracting 100 000 assets, you might want to
            limit the extraction to 1000 assets to test the setup.
        unpack_metadata (bool, optional): Whether to unpack metadata. Defaults to False, which yields the metadata as
            a JSON string.
        skip_metadata_values (set[str] | frozenset[str] | None, optional): A set of values to skip when unpacking
            metadata. Defaults to frozenset({"nan", "null", "none", ""}).
    """

    _default_rdf_type = "DataSet"

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: SequenceNotStr[str],
        namespace: Namespace | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        return cls(
            client.data_sets.retrieve_multiple(external_ids=data_set_external_id),
            namespace=namespace,
            total=len(data_set_external_id),
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )

    @classmethod
    def from_file(
        cls,
        file_path: str,
        namespace: Namespace | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        data_sets = DataSetList.load(Path(file_path).read_text())
        return cls(
            data_sets,
            namespace=namespace,
            total=len(data_sets),
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )

    def _item2triples(self, item: DataSet) -> list[Triple]:
        """Converts an asset to triples."""
        id_ = self.namespace[f"{InstanceIdPrefix.data_set}{item.id}"]

        type_ = self._get_rdf_type(item)

        triples: list[Triple] = [(id_, RDF.type, self.namespace[type_])]

        # Create attributes
        if item.name:
            triples.append((id_, self.namespace.name, Literal(item.name)))

        if item.description:
            triples.append((id_, self.namespace.description, Literal(item.description)))

        if item.external_id:
            triples.append((id_, self.namespace.external_id, Literal(item.external_id)))

        # properties' ref creation and update
        triples.append(
            (
                id_,
                self.namespace.created_time,
                Literal(datetime.fromtimestamp(item.created_time / 1000, timezone.utc)),
            )
        )
        triples.append(
            (
                id_,
                self.namespace.last_updated_time,
                Literal(datetime.fromtimestamp(item.last_updated_time / 1000, timezone.utc)),
            )
        )

        if item.write_protected:
            triples.append((id_, self.namespace.write_protected, Literal(item.write_protected)))

        if item.metadata:
            triples.extend(self._metadata_to_triples(id_, item.metadata))

        return triples



from pathlib import Path

from cognite.client.data_classes.data_modeling.ids import DataModelId
from rdflib import DCTERMS, OWL, RDF, RDFS, SKOS, XSD, Namespace, URIRef

from cognite import neat


def _is_in_notebook() -> bool:
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def _is_in_browser() -> bool:
    try:
        from pyodide.ffi import IN_BROWSER  # type: ignore [import-not-found]
    except ModuleNotFoundError:
        return False
    return IN_BROWSER


IN_PYODIDE = _is_in_browser()
IN_NOTEBOOK = _is_in_notebook() or IN_PYODIDE


PACKAGE_DIRECTORY = Path(neat.__file__).parent
COGNITE_MODELS = (
    DataModelId("cdf_cdm", "CogniteCore", "v1"),
    DataModelId("cdf_idm", "CogniteProcessIndustries", "v1"),
)
DMS_LISTABLE_PROPERTY_LIMIT = 1000

EXAMPLE_RULES = PACKAGE_DIRECTORY / "_rules" / "examples"
EXAMPLE_GRAPHS = PACKAGE_DIRECTORY / "_graph" / "examples"
EXAMPLE_WORKFLOWS = PACKAGE_DIRECTORY / "_workflows" / "examples"

DEFAULT_SPACE_URI = "http://purl.org/cognite/{space}#"
DEFAULT_NAMESPACE = Namespace("http://purl.org/cognite/neat#")
DEFAULT_BASE_URI = URIRef(DEFAULT_NAMESPACE)
CLASSIC_CDF_NAMESPACE = Namespace("http://purl.org/cognite/cdf-classic#")
UNKNOWN_TYPE = DEFAULT_NAMESPACE.UnknownType


def get_default_prefixes() -> dict[str, Namespace]:
    return {
        "rdf": RDF._NS,
        "rdfs": RDFS._NS,
        "dct": DCTERMS._NS,
        "skos": SKOS._NS,
        "owl": OWL._NS,
        "xsd": XSD._NS,
        "pav": Namespace("http://purl.org/pav/"),
    }


DEFAULT_URI = ""

DEFAULT_DOCS_URL = "https://cognite-neat.readthedocs-hosted.com/en/latest/"

DMS_CONTAINER_PROPERTY_SIZE_LIMIT = 100
DMS_VIEW_CONTAINER_SIZE_LIMIT = 10



from dataclasses import dataclass, field
from typing import Literal, cast

from cognite.neat._issues import IssueList
from cognite.neat._rules._shared import ReadRules, VerifiedRules
from cognite.neat._rules.models.dms._rules import DMSRules
from cognite.neat._rules.models.information._rules import InformationRules
from cognite.neat._rules.models.information._rules_input import InformationInputRules
from cognite.neat._store import NeatGraphStore
from cognite.neat._utils.upload import UploadResultList

from .exceptions import NeatSessionError


@dataclass
class SessionState:
    store_type: Literal["memory", "oxigraph"]
    input_rules: list[ReadRules] = field(default_factory=list)
    verified_rules: list[VerifiedRules] = field(default_factory=list)
    issue_lists: list[IssueList] = field(default_factory=list)
    outcome: list[UploadResultList] = field(default_factory=list)
    _store: NeatGraphStore | None = field(init=False, default=None)

    @property
    def store(self) -> NeatGraphStore:
        if not self.has_store:
            if self.store_type == "oxigraph":
                self._store = NeatGraphStore.from_oxi_store()
            else:
                self._store = NeatGraphStore.from_memory_store()
        return cast(NeatGraphStore, self._store)

    @property
    def input_rule(self) -> ReadRules:
        if not self.input_rules:
            raise NeatSessionError("No input data model available. Try using [bold].read[/bold] to load a data model.")
        return self.input_rules[-1]

    @property
    def last_outcome(self) -> UploadResultList:
        if not self.outcome:
            raise NeatSessionError(
                "No outcome available. Try using [bold].to.cdf[/bold] to upload a data model/instances."
            )
        return self.outcome[-1]

    @property
    def information_input_rule(self) -> ReadRules | None:
        if self.input_rules:
            for rule in self.input_rules[::-1]:
                if isinstance(rule.rules, InformationInputRules):
                    return rule
        return None

    @property
    def last_verified_rule(self) -> VerifiedRules:
        if not self.verified_rules:
            raise NeatSessionError(
                "No data model available to verify. Try using  [bold].read[/bold] to load a data model."
            )
        return self.verified_rules[-1]

    @property
    def last_verified_dms_rules(self) -> DMSRules:
        if self.verified_rules:
            for rules in self.verified_rules[::-1]:
                if isinstance(rules, DMSRules):
                    return rules

        raise NeatSessionError(
            'No verified DMS data model. Try using  [bold].convert("DMS")[/bold]'
            " to convert verified information model to verified DMS model."
        )

    @property
    def last_verified_information_rules(self) -> InformationRules:
        if self.verified_rules:
            for rules in self.verified_rules[::-1]:
                if isinstance(rules, InformationRules):
                    return rules

        raise NeatSessionError(
            "No verified information data model. Try using  [bold].verify()[/bold]"
            " to convert unverified information model to verified information model."
        )

    @property
    def has_store(self) -> bool:
        return self._store is not None

    @property
    def has_verified_rules(self) -> bool:
        return bool(self.verified_rules)

    @property
    def last_issues(self) -> IssueList:
        if not self.issue_lists:
            raise NeatSessionError("No issues available. Try using [bold].verify()[/bold] to verify a data model.")
        return self.issue_lists[-1]



from ._base import RulesPipeline, RulesTransformer
from ._converters import (
    AssetToInformation,
    ConvertToRules,
    DMSToInformation,
    InformationToAsset,
    InformationToDMS,
    ReduceCogniteModel,
    SetIDDMSModel,
    ToCompliantEntities,
    ToExtension,
)
from ._mapping import MapOneToOne, RuleMapper
from ._pipelines import ImporterPipeline
from ._verification import VerifyAnyRules, VerifyAssetRules, VerifyDMSRules, VerifyInformationRules

__all__ = [
    "ImporterPipeline",
    "RulesTransformer",
    "RulesPipeline",
    "InformationToDMS",
    "InformationToAsset",
    "ConvertToRules",
    "AssetToInformation",
    "DMSToInformation",
    "VerifyAssetRules",
    "VerifyDMSRules",
    "VerifyInformationRules",
    "VerifyAnyRules",
    "MapOneToOne",
    "ToCompliantEntities",
    "RuleMapper",
    "ToExtension",
    "ReduceCogniteModel",
    "SetIDDMSModel",
]



from typing import cast

from rdflib import Graph

from cognite.neat._rules.importers._rdf._shared import (
    clean_up_classes,
    make_classes_compliant,
    parse_raw_classes_dataframe,
)


def parse_imf_to_classes(graph: Graph, language: str = "en") -> list[dict]:
    """Parse IMF elements from RDF-graph and extract classes to pandas dataframe.

    Args:
        graph: Graph containing imf elements
        language: Language to use for parsing, by default "en"

    Returns:
        Dataframe containing imf elements

    !!! note "IMF Compliance"
        The IMF elements are expressed in RDF, primarily using SHACL and OWL. To ensure
        that the resulting classes are compliant with CDF, similar validation checks as
        in the OWL ontology importer are applied.

        For the IMF-types more of the compliance logic is placed directly in the SPARQL
        query. Among these are the creation of class name not starting with a number,
        and ensuring that all classes have a parent.

        IMF-attributes are considered both classes and properties. This kind of punning
        is necessary to capture additional information carried by attributes. They carry,
        among other things, a set of relationsships to reference terms, units of measure,
        and qualifiers that together make up the meaning of the attribute.
    """

    query = """
    SELECT ?class ?name ?description ?parentClass ?reference ?match ?comment
    WHERE {
        # Finding IMF - elements
        VALUES ?type { imf:BlockType imf:TerminalType imf:AttributeType }
        ?imfClass a ?type .
        OPTIONAL {?imfClass rdfs:subClassOf ?parent }.
        OPTIONAL {?imfClass rdfs:label | skos:prefLabel ?name }.

        # Note: Bug in PCA has lead to the use non-existing term skos:description. This will be replaced
        # with the correct skos:definition in the near future, so both terms are included here.
        OPTIONAL {?imfClass rdfs:comment | skos:definition | skos:description ?description} .

        # Finding the last segment of the class IRI
        BIND(STR(?imfClass) AS ?classString)
        BIND(REPLACE(?classString, "^.*[/#]([^/#]*)$", "$1") AS ?tempSegment)
        BIND(REPLACE(?tempSegment, "-", "_") AS ?classSegment)
        BIND(IF(CONTAINS(?classString, "imf/"), CONCAT("IMF_", ?classSegment) , ?classSegment) AS ?class)

        # Add imf:Attribute as parent class
        BIND(IF(!bound(?parent) && ?type = imf:AttributeType, imf:Attribute, ?parent) AS ?parentClass)

        # Rebind the IRI of the IMF-type to the ?reference variable to align with dataframe column headers
        # This is solely for readability, the ?imfClass could have been returned directly instead of ?reference
        BIND(?imfClass AS ?reference)

        FILTER (!isBlank(?class))
        FILTER (!bound(?parentClass) || !isBlank(?parentClass))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "en"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
    }
    """

    # create raw dataframe
    raw_df = parse_raw_classes_dataframe(cast(list[tuple], list(graph.query(query.replace("en", language)))))
    if raw_df.empty:
        return []

    # group values and clean up
    processed_df = clean_up_classes(raw_df)

    # make compliant
    processed_df = make_classes_compliant(processed_df, importer="IMF")

    # Make Parent Class list elements into string joined with comma
    processed_df["Parent Class"] = processed_df["Parent Class"].apply(
        lambda x: ", ".join(x) if isinstance(x, list) and x else None
    )

    return processed_df.dropna(axis=0, how="all").replace(float("nan"), None).to_dict(orient="records")



from pathlib import Path

from rdflib import DC, DCTERMS, OWL, RDF, RDFS, SH, SKOS, XSD, Graph, Namespace, URIRef

from cognite.neat._issues import IssueList
from cognite.neat._issues.errors import FileReadError
from cognite.neat._rules._shared import ReadRules
from cognite.neat._rules.importers._base import BaseImporter
from cognite.neat._rules.models.data_types import AnyURI
from cognite.neat._rules.models.entities import UnknownEntity
from cognite.neat._rules.models.information import (
    InformationInputRules,
)
from cognite.neat._store import NeatGraphStore
from cognite.neat._utils.rdf_ import get_namespace, uri_to_short_form

DEFAULT_NON_EXISTING_NODE_TYPE = AnyURI()


class BaseRDFImporter(BaseImporter[InformationInputRules]):
    """Baser RDF importers used for all rules importers that are using RDF as input.

    Args:
        issue_list: Issue list to store issues
        graph: Knowledge graph
        prefix: Prefix to be used for the imported rules
    """

    def __init__(
        self,
        issue_list: IssueList,
        graph: Graph,
        prefix: str,
        max_number_of_instance: int,
        non_existing_node_type: UnknownEntity | AnyURI,
    ) -> None:
        self.issue_list = issue_list
        self.graph = graph
        self.prefix = prefix
        self.max_number_of_instance = max_number_of_instance
        self.non_existing_node_type = non_existing_node_type

    @classmethod
    def from_graph_store(
        cls,
        store: NeatGraphStore,
        prefix: str = "neat",
        max_number_of_instance: int = -1,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
    ):
        return cls(
            IssueList(title=f"{cls.__name__} issues"),
            store.graph,
            prefix=prefix,
            max_number_of_instance=max_number_of_instance,
            non_existing_node_type=non_existing_node_type,
        )

    @classmethod
    def from_file(
        cls,
        filepath: Path,
        prefix: str = "neat",
        max_number_of_instance: int = -1,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
    ):
        issue_list = IssueList(title=f"{cls.__name__} issues")

        graph = Graph()
        try:
            graph.parse(filepath)
        except Exception as e:
            issue_list.append(FileReadError(filepath, str(e)))

        # bind key namespaces
        graph.bind("owl", OWL)
        graph.bind("rdf", RDF)
        graph.bind("rdfs", RDFS)
        graph.bind("dcterms", DCTERMS)
        graph.bind("dc", DC)
        graph.bind("skos", SKOS)
        graph.bind("sh", SH)
        graph.bind("xsd", XSD)
        graph.bind("imf", "http://ns.imfid.org/imf#")

        return cls(
            issue_list,
            graph,
            prefix=prefix,
            max_number_of_instance=max_number_of_instance,
            non_existing_node_type=non_existing_node_type,
        )

    def to_rules(
        self,
    ) -> ReadRules[InformationInputRules]:
        """
        Creates `Rules` object from the data for target role.
        """

        if self.issue_list.has_errors:
            # In case there were errors during the import, the to_rules method will return None
            return ReadRules(None, self.issue_list, {})

        rules_dict = self._to_rules_components()

        rules = InformationInputRules.load(rules_dict)

        self._add_transformation(rules)

        return ReadRules(rules, self.issue_list, {})

    def _to_rules_components(self) -> dict:
        raise NotImplementedError()

    @classmethod
    def _add_transformation(cls, rules: InformationInputRules) -> None:
        rules.prefixes = {}

        rules.prefixes[rules.metadata.prefix] = Namespace(rules.metadata.namespace)

        class_ref_dict = {}
        for class_ in rules.classes:
            if class_.reference:
                cls._add_uri_namespace_to_prefixes(URIRef(class_.reference), rules.prefixes)
                class_ref_dict[class_.class_] = f"{uri_to_short_form(URIRef(class_.reference), rules.prefixes)}"

        for property_ in rules.properties:
            if property_.reference:
                cls._add_uri_namespace_to_prefixes(URIRef(property_.reference), rules.prefixes)
                if class_id := class_ref_dict.get(property_.class_, None):
                    property_id = f"{uri_to_short_form(URIRef(property_.reference), rules.prefixes)}"
                    property_.transformation = f"{class_id}({property_id})"

    @classmethod
    def _add_uri_namespace_to_prefixes(cls, URI: URIRef, prefixes: dict[str, Namespace]):
        """Add URI to prefixes dict if not already present

        Args:
            URI: URI from which namespace is being extracted
            prefixes: Dict of prefixes and namespaces
        """
        if Namespace(get_namespace(URI)) not in prefixes.values():
            prefixes[f"prefix-{len(prefixes)+1}"] = Namespace(get_namespace(URI))



from abc import ABC
from collections import defaultdict

from cognite.neat._rules._shared import JustRules, OutRules
from cognite.neat._rules.models import DMSRules, InformationRules
from cognite.neat._rules.models._base_rules import ClassRef
from cognite.neat._rules.models.dms import DMSProperty
from cognite.neat._rules.models.entities import ClassEntity, ReferenceEntity
from cognite.neat._rules.models.information import InformationClass
from cognite.neat._rules.models.mapping import RuleMapping

from ._base import RulesTransformer


class MapOntoTransformers(RulesTransformer[DMSRules, DMSRules], ABC):
    """Base class for transformers that map one rule onto another."""

    ...


class MapOneToOne(MapOntoTransformers):
    """Takes transform data models and makes it into an extension of the reference data model.

    Note this transformer mutates the input rules.

    The argument view_extension_mapping is a dictionary where the keys are views of this data model,
    and each value is the view of the reference data model that the view should extend. For example:

    ```python
    view_extension_mapping = {"Pump": "Asset"}
    ```

    This would make the view "Pump" in this data model extend the view "Asset" in the reference data model.
    Note that all the keys in the dictionary must be external ids of views in this data model,
    and all the values must be external ids of views in the reference data model.

    Args:
        reference: The reference data model
        view_extension_mapping: A dictionary mapping views in this data model to views in the reference data model
        default_extension: The default view in the reference data model that views in this
            data model should extend if no mapping is provided.

    """

    def __init__(
        self, reference: DMSRules, view_extension_mapping: dict[str, str], default_extension: str | None = None
    ) -> None:
        self.reference = reference
        self.view_extension_mapping = view_extension_mapping
        self.default_extension = default_extension

    def transform(self, rules: DMSRules | OutRules[DMSRules]) -> JustRules[DMSRules]:
        solution: DMSRules = self._to_rules(rules)
        if solution.reference is not None:
            raise ValueError("Reference already exists")
        solution.reference = self.reference
        view_by_external_id = {view.view.external_id: view for view in solution.views}
        ref_view_by_external_id = {view.view.external_id: view for view in self.reference.views}

        if invalid_views := set(self.view_extension_mapping.keys()) - set(view_by_external_id.keys()):
            raise ValueError(f"Views are not in this dat model {invalid_views}")
        if invalid_views := set(self.view_extension_mapping.values()) - set(ref_view_by_external_id.keys()):
            raise ValueError(f"Views are not in the reference data model {invalid_views}")
        if self.default_extension and self.default_extension not in ref_view_by_external_id:
            raise ValueError(f"Default extension view not in the reference data model {self.default_extension}")

        properties_by_view_external_id: dict[str, dict[str, DMSProperty]] = defaultdict(dict)
        for prop in solution.properties:
            properties_by_view_external_id[prop.view.external_id][prop.view_property] = prop

        ref_properties_by_view_external_id: dict[str, dict[str, DMSProperty]] = defaultdict(dict)
        for prop in self.reference.properties:
            ref_properties_by_view_external_id[prop.view.external_id][prop.view_property] = prop

        for view_external_id, view in view_by_external_id.items():
            if view_external_id in self.view_extension_mapping:
                ref_external_id = self.view_extension_mapping[view_external_id]
            elif self.default_extension:
                ref_external_id = self.default_extension
            else:
                continue

            ref_view = ref_view_by_external_id[ref_external_id]
            shared_properties = set(properties_by_view_external_id[view_external_id].keys()) & set(
                ref_properties_by_view_external_id[ref_external_id].keys()
            )
            if shared_properties:
                if view.implements is None:
                    view.implements = [ref_view.view]
                elif isinstance(view.implements, list) and ref_view.view not in view.implements:
                    view.implements.append(ref_view.view)
            for prop_name in shared_properties:
                prop = properties_by_view_external_id[view_external_id][prop_name]
                ref_prop = ref_properties_by_view_external_id[ref_external_id][prop_name]
                if ref_prop.container and ref_prop.container_property:
                    prop.container = ref_prop.container
                    prop.container_property = ref_prop.container_property
                prop.reference = ReferenceEntity.from_entity(ref_prop.view, ref_prop.view_property)

        return JustRules(solution)


class RuleMapper(RulesTransformer[InformationRules, InformationRules]):
    """Maps properties and classes using the given mapping.

    **Note**: This transformer mutates the input rules.

    Args:
        mapping: The mapping to use.

    """

    def __init__(self, mapping: RuleMapping) -> None:
        self.mapping = mapping

    def transform(self, rules: InformationRules | OutRules[InformationRules]) -> JustRules[InformationRules]:
        input_rules = self._to_rules(rules)

        destination_by_source = self.mapping.properties.as_destination_by_source()
        destination_cls_by_source = self.mapping.classes.as_destination_by_source()
        used_destination_classes: set[ClassEntity] = set()
        for prop in input_rules.properties:
            if destination_prop := destination_by_source.get(prop.as_reference()):
                prop.class_ = destination_prop.class_
                prop.property_ = destination_prop.property_
                used_destination_classes.add(destination_prop.class_)
            elif destination_cls := destination_cls_by_source.get(ClassRef(Class=prop.class_)):
                # If the property is not in the mapping, but the class is,
                # then we should map the class to the destination
                prop.class_ = destination_cls.class_

        for cls_ in input_rules.classes:
            if destination_cls := destination_cls_by_source.get(cls_.as_reference()):
                cls_.class_ = destination_cls.class_
        existing_classes = {cls_.class_ for cls_ in input_rules.classes}
        for new_cls in used_destination_classes - existing_classes:
            input_rules.classes.append(InformationClass(class_=new_cls))

        return JustRules(input_rules)



import json
import re
import sys
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Set
from typing import Generic, TypeVar

from cognite.client.data_classes._base import CogniteResource
from rdflib import XSD, Literal, Namespace, URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._graph.extractors._base import BaseExtractor
from cognite.neat._graph.models import Triple
from cognite.neat._utils.auxiliary import string_to_ideal_type

T_CogniteResource = TypeVar("T_CogniteResource", bound=CogniteResource)

DEFAULT_SKIP_METADATA_VALUES = frozenset({"nan", "null", "none", ""})

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


class InstanceIdPrefix(StrEnum):
    asset = "Asset_"
    label = "Label_"
    relationship = "Relationship_"
    sequence = "Sequence_"
    file = "File_"
    time_series = "TimeSeries_"
    event = "Event_"
    data_set = "DataSet_"

    @classmethod
    def from_str(cls, raw: str) -> "InstanceIdPrefix":
        raw = raw.title() + "_"
        if raw == "Timeseries_":
            return cls.time_series
        else:
            return cls(raw)


class ClassicCDFBaseExtractor(BaseExtractor, ABC, Generic[T_CogniteResource]):
    """This is the Base Extractor for all classic CDF resources.

    A classic resource is recognized in that it has a metadata attribute of type dict[str, str].

    Args:
        items (Iterable[T_CogniteResource]): An iterable of classic resource.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
        to_type (Callable[[T_CogniteResource], str | None], optional): A function to convert an item to a type.
            Defaults to None. If None or if the function returns None, the asset will be set to the default type.
        total (int, optional): The total number of items to load. If passed, you will get a progress bar if rich
            is installed. Defaults to None.
        limit (int, optional): The maximal number of items to load. Defaults to None. This is typically used for
            testing setup of the extractor. For example, if you are extracting 100 000 assets, you might want to
            limit the extraction to 1000 assets to test the setup.
        unpack_metadata (bool, optional): Whether to unpack metadata. Defaults to False, which yields the metadata as
            a JSON string.
        skip_metadata_values (set[str] | frozenset[str] | None, optional): If you are unpacking metadata, then
           values in this set will be skipped.
    """

    _default_rdf_type: str
    _SPACE_PATTERN = re.compile(r"\s+")

    def __init__(
        self,
        items: Iterable[T_CogniteResource],
        namespace: Namespace | None = None,
        to_type: Callable[[T_CogniteResource], str | None] | None = None,
        total: int | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.items = items
        self.to_type = to_type
        self.total = total
        self.limit = min(limit, total) if limit and total else limit
        self.unpack_metadata = unpack_metadata
        self.skip_metadata_values = skip_metadata_values

    def extract(self) -> Iterable[Triple]:
        """Extracts an asset with the given asset_id."""
        if self.total:
            try:
                from rich.progress import track
            except ModuleNotFoundError:
                to_iterate = self.items
            else:
                to_iterate = track(
                    self.items,
                    total=self.limit or self.total,
                    description=f"Extracting {type(self).__name__.removesuffix('Extractor')}",
                )
        else:
            to_iterate = self.items
        for no, asset in enumerate(to_iterate):
            yield from self._item2triples(asset)
            if self.limit and no >= self.limit:
                break

    @abstractmethod
    def _item2triples(self, item: T_CogniteResource) -> list[Triple]:
        raise NotImplementedError()

    def _metadata_to_triples(self, id_: URIRef, metadata: dict[str, str]) -> Iterable[Triple]:
        if self.unpack_metadata:
            for key, value in metadata.items():
                if value and (self.skip_metadata_values is None or value.casefold() not in self.skip_metadata_values):
                    yield (
                        id_,
                        self.namespace[key],
                        Literal(string_to_ideal_type(value)),
                    )
        else:
            yield id_, self.namespace.metadata, Literal(json.dumps(metadata), datatype=XSD._NS["json"])

    def _get_rdf_type(self, item: T_CogniteResource) -> str:
        type_ = self._default_rdf_type
        if self.to_type:
            type_ = self.to_type(item) or type_
        return self._SPACE_PATTERN.sub("_", type_)



import json
from pathlib import Path
from typing import ClassVar, cast

from cognite.neat._graph.extractors import RdfFileExtractor
from cognite.neat._graph.extractors._mock_graph_generator import MockGraphGenerator
from cognite.neat._issues.errors import WorkflowStepNotInitializedError
from cognite.neat._rules._shared import DMSRules, InformationRules
from cognite.neat._workflows.model import FlowMessage, StepExecutionStatus
from cognite.neat._workflows.steps.data_contracts import MultiRuleData, NeatGraph
from cognite.neat._workflows.steps.step_model import Configurable, Step

__all__ = ["GraphFromRdfFile", "GraphFromMockData"]

CATEGORY = __name__.split(".")[-1].replace("_", " ").title()


class GraphFromMockData(Step):
    """
    This step generate mock graph based on the defined classes and target number of instances
    """

    description = "This step generate mock graph based on the defined classes and target number of instances"
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="Class count",
            value="",
            label="Target number of instances for each class",
        ),
        Configurable(
            name="Graph",
            value="solution",
            label="The name of target graph.",
            options=["source", "solution"],
        ),
    ]

    def run(  # type: ignore[override, syntax]
        self, rules: MultiRuleData, store: NeatGraph
    ) -> FlowMessage:
        if self.configs is None:
            raise WorkflowStepNotInitializedError(type(self).__name__)

        if not rules.information and not rules.dms:
            return FlowMessage(
                error_text="Rules must be made either by Information Architect or DMS Architect!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        try:
            class_count = json.loads(self.configs["Class count"]) if self.configs["Class count"] else {}
        except Exception:
            return FlowMessage(
                error_text="Defected JSON stored in class_count",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        extractor = MockGraphGenerator(
            cast(InformationRules | DMSRules, rules.information or rules.dms),
            class_count,
        )

        store.graph.write(extractor)

        return FlowMessage(output_text=f"Instances loaded to the {store.__class__.__name__}")


class GraphFromRdfFile(Step):
    """
    This step extract instances from a file into the source graph. The file must be in RDF format.
    """

    description = "This step extract instances from a file into the source graph. The file must be in RDF format."
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="File path",
            value="source-graphs/source-graph-dump.xml",
            label="File name of source graph data dump in RDF format",
        )
    ]

    def run(self, store: NeatGraph) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise WorkflowStepNotInitializedError(type(self).__name__)

        if source_file := self.configs["File path"]:
            store.graph.write(
                RdfFileExtractor(  # type: ignore[abstract]
                    filepath=self.data_store_path / Path(source_file),
                )
            )

        else:
            raise ValueError("You need a valid file path to be specified")

        return FlowMessage(output_text="Instances loaded to NeatGraph!")



from typing import cast

from rdflib import Graph

from cognite.neat._rules.importers._rdf._shared import (
    clean_up_properties,
    make_properties_compliant,
    parse_raw_properties_dataframe,
)


def parse_owl_properties(graph: Graph, language: str = "en") -> list[dict]:
    """Parse owl properties from graph to pandas dataframe.

    Args:
        graph: Graph containing owl properties
        language: Language to use for parsing, by default "en"

    Returns:
        List of dictionaries containing owl properties
    """

    query = """

    SELECT ?class ?property ?name ?description ?type ?minCount ?maxCount ?default ?reference
     ?match ?comment ?propertyType
    WHERE {
        ?property a ?propertyType.
        FILTER (?propertyType IN (owl:ObjectProperty, owl:DatatypeProperty ) )
        OPTIONAL {?property rdfs:domain ?class }.
        OPTIONAL {?property rdfs:range ?type }.
        OPTIONAL {?property rdfs:label ?name }.
        OPTIONAL {?property rdfs:comment ?description} .
        OPTIONAL {?property owl:maxCardinality ?maxCount} .
        OPTIONAL {?property owl:minCardinality ?minCount} .
        FILTER (!isBlank(?property))
        FILTER (!bound(?type) || !isBlank(?type))
        FILTER (!bound(?class) || !isBlank(?class))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "en"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
        BIND(IF(bound(?minCount), ?minCount, 0) AS ?minCount)
        BIND(IF(bound(?maxCount), ?maxCount, 1) AS ?maxCount)
        BIND(?property AS ?reference)
    }
    """

    raw_df = parse_raw_properties_dataframe(cast(list[tuple], list(graph.query(query.replace("en", language)))))
    if raw_df.empty:
        return []

    # group values and clean up
    processed_df = clean_up_properties(raw_df)

    # make compliant
    processed_df = make_properties_compliant(processed_df, importer="OWL")

    # drop column _property_type, which was a helper column:
    processed_df.drop(columns=["_property_type"], inplace=True)

    return processed_df.to_dict(orient="records")
